import time
import random

def begin(): #used to restart game in order to reset inventory etc.
    game = Engine(name)
    game.load()
    game.play()

def linebreak(text, col=65):
    pretty_line = ""
    count = 0
    for char in text:
        if count > col and char == " ":
            pretty_line += "\n"
            count = 0
        else:
            pretty_line += char
            count += 1
    return pretty_line

class Node:
        def __init__(self, name, description):
                self.name = name
                self.description = description
                self.edges = []
                self.objects = []
                self.printed = False #meant to prevent duplicate printing of a node description when objects are picked up etc.

class Edge:
        def __init__(self, name, from_node, commands, to_node, description, health, constraints=None):
                self.name = name
                self.from_node = from_node
                self.commands = commands
                self.to_node = to_node
                self.description = description
                self.hp = health
                self.constraints = constraints
                self.locked = False #used for constrained edges

class Object:
        def __init__(self, location, name, shrt_desc, lng_desc, weight):
                self.location = location
                self.name = name
                self.shrt_desc = shrt_desc
                self.lng_desc = lng_desc
                self.weight = weight

class Constraint:
        def __init__(self, name, second, third): #obj_name and obj_location were switched out for second and third
            #because eval and exec do not fit into these descriptions and writing the code for constraints
            #was far easier when approaching them this way.
                self.name = name
                self.second = second
                self.third = third
                self.met = False
                
class Greeting:
    def __init__(self, description):
        self.description = description

class Cheat:
    def __init__(self, name, code, result): #Cheat names must contain 'cheat' for constraint purposes
        self.name = name
        self.code = code
        self.result = result

class Inventory: #backpack
        def __init__(self):
                self.objects = []
                self.cheats = []
                self.weight = 0

class Engine:
        def __init__(self, filename):
                self.filename = filename
                self.initial_node = None
                self.initial_time = time.time() #is used to calculate current time on the edges in case
                #anyone is implementing time in their code for constraints
                self.health = 10
                self.counter = 0
                self.backpack = False #allows for game files where backpack::max weight::VALUE does not exist
                self.max_weight = 0
                self.greeting = None
                self.constraints = []
                self.inventory = Inventory()
                self.node_commands = [] #lists possible edge commands to prevent duplications and duplicate edge accesses
                self.not_print = [] #prevents descriptions of locked edges with duplicate commands from appearing and giving stuff away
                self.command_used = False #prevents two edges from being accessed if they have the same command and one is limited by
                #a met constraint while the other is not.
                self.cheats = [] #list of cheats because they're not determined by node or object etc
                self.cheat_count = 0 #will kill a cheater
        def load(self):
                self.nodes = {}
                edges = []
                objects = []
                fle = open(self.filename)
                for line in fle:
                        if len(line) == 0 or line[0] == '#':
                                continue
                        items = line.strip().split('::')
                        if items[0] == 'node':
                                n = Node(*items[1:])
                                if self.initial_node is None:
                                        self.initial_node = n
                                self.nodes[items[1]] = n
                        elif items[0] == 'edge':
                                edges.append(Edge(*items[1:]))
                        elif items[0] == 'object':
                                objects.append(Object(*items[1:]))
                        elif items[0] == 'backpack':
                            self.backpack = True
                            self.max_weight = int(items[2])
                        elif items[0] == 'greeting':
                            g = Greeting(*items[1:])
                            if self.greeting is None:
                                self.greeting = g
                        elif items[0] == 'constraint':
                               self.constraints.append(Constraint(*items[1:]))
                        elif items[0] == 'cheat':
                            self.cheats.append(Cheat(*items[1:]))
                for edge in edges:
                        self.nodes[edge.from_node].edges.append(edge)
                for obj in objects:
                        self.nodes[obj.location].objects.append(obj)
        #Clears lists keeping track of commands
        def empty(self):
            self.nodes_commands = []
            self.not_print = []
        #Testing constraints
        def is_met(self, constraint):
            if constraint.second == 'eval':
                if eval(constraint.third) is True:
                    constraint.met = True
                    return constraint.met
            elif constraint.second == 'exec':
                constraint.met = True
                exec(constraint.third)
            elif 'cheat' in constraint.second:
                for cheat in self.inventory.cheats:
                    if cheat.name in constraint.second:
                        constraint.met = True
                        return constraint.met
            else:
                if constraint.third == 'inventory' or constraint.third == 'backpack':
                    objects = constraint.second.split(', ')
                    for obj in self.inventory.objects:
                        if obj.name in objects:
                            constraint.met = True
                            return constraint.met
                elif constraint.third in self.nodes:
                    for obj in self.nodes[constraint.third].objects:
                        if obj in self.nodes[constraint.third].objects:
                            constraint.met = True
                            return constraint.met
        def play(self):
            current_node = self.initial_node
            if self.greeting is not None:
                print('\n',linebreak(self.greeting.description), '\n')
            while True:
                if current_node.printed is False:
                    current_node.printed = True
                    print(linebreak(current_node.description))
                command = input("command>> ")
                c = command.split()
                self.command_used = False
                for edge in current_node.edges:
                    if command in edge.commands:
                        if edge.constraints is not None:
                            edge.locked = True
                            for constraint in self.constraints:
                                if constraint.name in edge.constraints:
                                    self.is_met(constraint)
                                    if constraint.met is True: #constraint is met
                                        edge.locked = False #unlocks edge
                                        current_node.printed = False
                                        current_node = self.nodes[edge.to_node]
                                        self.counter += 1
                                        self.health += int(edge.hp)
                                        timer = time.time() - self.initial_time
                                        self.empty()
                                        constraint.met = False #edge is re-locked
                                        if self.health > 0 and edge.locked is False:
                                            print(linebreak(edge.description))
                                            self.command_used = True
                                    elif constraint.met is False:
                                        for n in self.nodes[edge.from_node].edges:
                                            self.node_commands.append(n.commands)
                                        for i in self.node_commands:
                                            if self.node_commands.count(i) > 1:
                                                self.not_print.append(i)
                                                self.node_commands.remove(i)
                                        for i in self.not_print:
                                            if (str(command) in i) is True:
                                                edge.locked = True
                                                self.empty()
                                            else:
                                                edge.locked = True
                                                print("You can't do that yet, ", user+".")
                                                self.empty()
                        elif edge.constraints is None and self.command_used is False:
                            current_node.printed = False
                            current_node = self.nodes[edge.to_node]
                            self.counter += 1
                            self.health += int(edge.hp)
                            timer = time.time() - self.initial_time
                            if self.health > 0 and edge.locked is False:
                                print(linebreak(edge.description))
                    if self.health <= 0:
                        print("Your health has reached zero. Game Over, ", user+".")
                        ans = input("Play again? [y/n]>> ")
                        if ans == 'y':
                            self.health = 10
                            begin()
                        else:
                            print("Thanks for playing,", user,"!")
                        break
                        return
                    
                #Actions for objects in nodes
                for obj in current_node.objects:
                    if c[0] == 'get':
                        if c[1] in obj.name:
                            if self.backpack is True:
                                if (self.inventory.weight + int(obj.weight)) < self.max_weight:
                                    self.inventory.objects.append(obj)
                                    current_node.objects.remove(obj)
                                    self.inventory.weight += int(obj.weight)
                                    print('You got the', c[1])
                                else:
                                    print(user+', you cannot carry more weight')
                            else:
                                self.inventory.objects.append(obj)
                                current_node.objects.remove(obj)
                                print('You got the', c[1])
                                
                #Inventory and/or backpack actions
                if 'inventory' in command or 'backpack' in command:
                    if len(self.inventory.cheats) != 0:
                        ans = input("Which would you like to view? -- <cheats> or <items>? ")
                        if "cheats" in ans:
                            print(user+"'s cheat(s)")
                            for i in range(len(self.inventory.cheats)):
                                print(self.inventory.cheats[i].name)
                        elif "items" in ans:
                            print(user,', you have the...')
                            for i in range(len(self.inventory.objects)):
                                    print(self.inventory.objects[i].name)
                            if self.backpack is True:
                                ans = input('Would you like to know the total weight of the items you are carrying? [y/n]>> ')
                                if 'y' in ans:
                                    print('current weight: ', self.inventory.weight)
                    else:
                        print(user, ', you have the...')
                        for i in range(len(self.inventory.objects)):
                            print(self.inventory.objects[i].name)
                        if self.backpack is True:
                            ans = input('Would you like to know the total weight of the items you are carrying? [y/n]>> ')
                            if 'y' in ans:
                                print('current weight: ', self.inventory.weight)
                                
                #Check health
                if 'health' in command:
                    print('Current health: ', self.health)
                    
                #Checking move counter
                if 'moves' in command:
                    print("Current moves you've made: ",self.counter)
                    
                #Responses to end of game.
                if 'GAME OVER' in current_node.description or 'EPIC WIN' in current_node.description or 'THE END' in current_node.description:
                        print(linebreak(current_node.description))
                        ans = input("Play again? [y/n]>> ")
                        if 'y' in ans:
                            begin()
                        else:
                            print("Thanks for playing,", user+"!")
                        break
                        return
                    
                #Help with edge commands
                if 'help' in command:
                    print("Here's some help,", user+'!')
                    for edge in game.nodes[current_node.name].edges:
                        print("you can", edge.commands.split(','), 'or')
                    print('you can <quit> the game.')
                    
                #Quitting the game
                if 'quit' in command:
                    print("Bye,", user+"!")
                    break
                    return
                
                #Start over
                if 'start over' in command:
                    ans = input("Are you sure you want to start over? You will lose all progress and items. [y/n]>> ")
                    if 'y' in ans:
                        print("Starting over.")
                        begin()
                    else:
                        break
                    return
                        
                #Responses to random input
                if "I'm scared" in command:
                    print("Too bad. :)", '\n')
                if "blow" and "up" in command or "Blow" and "up" in command:
                    print("With what? Your mind?", '\n')
                    
                #Cheating
                if command == 'cheat' or command == "Cheat":
                    if len(self.cheats) == 0: #if the game writer didn't include cheats
                        print("Cheater, cheater pumpkin eater! Don't try that too many times, or you'll run into trouble!")
                        self.cheat_count += 1
                        print(self.cheat_count)
                        if self.cheat_count == 3:
                            print("Like in the case of Little Bunny Fufu, you have disobeyed more than your fair share."+
                                  "Henceforth, you shall be a ghoul! And ghouls don't play games."+
                                  "Good bye, Ghoul", user+'.')
                            ans = input("Are you sorry for cheating? [y/n]>> ")
                            if 'y' in ans:
                                ans = input("Okay. Would you like to be a person again? [y/n]>> ")
                                if 'y' in ans:
                                    ans = input("*poof* You're human again! Would you like to play the game again? [y/n]>> ")
                                    if 'y' in ans:
                                        self.cheat_count = 0
                                        begin()
                                    else:
                                        print("Okay. Bye, Human", user+".")
                                        break
                                        return
                                else:
                                    print("Okay. Have fun as a smelly ghoul.")
                                    break
                                    return
                            else:
                                print("Fine. You can stay a ghoul for all I care.")
                                break
                                return
                    elif len(self.cheats) > 0: #if the game writer included cheats
                        print("So you think you can cheat, do you,", user+"? Let's see what you've got.")
                        query = input("Enter cheat code: ")
                        for i in self.cheats:
                            if query == i.code:
                                self.inventory.cheats.append(i)
                                print("This cheat was added to your inventory.")
                            else:
                                print("That's not the right code. Sorry.")
                                
                #Actions for objects in inventory
                for obj in self.inventory.objects:   
                    if c[0] == 'drop':
                        if c[1] in obj.name:
                            current_node.objects.append(obj)
                            self.inventory.objects.remove(obj)
                            print('You dropped the', obj.name)
                    if c[0] == 'view':
                        if c[1] in obj.name:
                            print(linebreak(obj.shrt_desc))
                    if c[0] == 'examine':
                        if c[1] in obj.name:
                            print(linebreak(obj.lng_desc))
                            
                #Action for cheats in inventory
                for cheat in self.inventory.cheats:
                    if c[0] == 'view':
                        if c[1] in cheat.name:
                            print(linebreak(cheat.result))


user = input("Enter your name: ")
print("Hello,", user+'.')
answer = input("Would you like to play a game?  [y/n]>> ")
if 'y' in answer or "Y" in answer:
    name = input("Please choose game file to run: ")
    game = Engine(name)
    game.load()
    game.play()    
else:
    print("All right. Maybe next time, then?")

